(** Time-stamp: <modified the 14/04/2008 (at 17:58) by Erwan Jahier> *)


open Predef
open SyntaxTreeCore
open CompiledData


(* exported *)
type 'a evaluator = 'a list list -> 'a list 
type typer = type_eff evaluator
type const_evaluator = const_eff evaluator
type clocker = clock_eff evaluator


(*********************************************************************************)
exception EvalConst_error of string
exception EvalType_error of string

let (type_error : type_eff list -> string -> 'a) =
  fun v expect -> 
    raise (EvalType_error(
	     "type mismatch"^(if expect = "" then "" else (expect^" expected"))))
      
let arity_error (v : 'a list) (expect : string) =
  raise (EvalType_error(
	   Printf.sprintf "arity error (%d arguments, whereas %s were expected)" 
	     (List.length v) expect))
    
(* exported *)
let (type_error_const : const_eff list -> string -> 'a) =
  fun v expect -> 
    raise (EvalConst_error(
	     "type mismatch"^(if expect = "" then "" else (expect^" expected"))))
      
(* exported *)
let arity_error_const (v : const_eff list) (expect : string) =
  raise (EvalConst_error(
	   Printf.sprintf "arity error (%d arguments, whereas %s were expected)" 
	     (List.length v) expect))
    
let not_evaluable op l =
  raise (EvalConst_error(
	   Printf.sprintf "The operator %s is not allowed in static expression" 
	     (op2string op)))
 

(*********************************************************************************)
(** 
   Naming rules. In the following function names,
     "a" stands for "any" or alpha (i.e., polymorph types)
     "b" stands for "bool"
     "i" stands for "int"
     "f" stands for "float"
     "o" stands for "overloaded", i.e.,  "int" ot "float"
     "s" stands for string

   Hence, for example, "iib" stands for "int -> int -> bool" 
*)
let (bbb_typer:typer) = fun ll -> match List.flatten ll with
  | [Bool_type_eff; Bool_type_eff] -> [Bool_type_eff]
  | [x0; x1] -> type_error [x0; x1] "bool*bool"
  | x -> arity_error x "2"
        
let (ooo_typer:typer) = fun ll -> match List.flatten ll with
  | [Int_type_eff; Int_type_eff] -> [Int_type_eff]
  | [Real_type_eff; Real_type_eff] -> [Real_type_eff]
  | [x0; x1] -> type_error [x0; x1] "int*int or real*real"
  | x ->  arity_error x "2"

let (iii_typer: typer) = fun ll -> match List.flatten ll with
  | [Int_type_eff; Int_type_eff] -> [Int_type_eff]
  | [x0; x1] -> (type_error [x0; x1] "int*int")
  | x        -> (arity_error x "2")
	
let (aab_typer: typer) = fun ll -> match List.flatten ll with 
  | [v0; v1] -> if v0 = v1 then [Bool_type_eff] else (type_error [v0; v1] "")
  | x -> (arity_error x "2")

let (fff_typer: typer) = fun ll -> match List.flatten ll with
  | [Real_type_eff; Real_type_eff] -> [Real_type_eff]
  | [x0;x1] -> (type_error [x0; x1] "real*real")
  | x    -> (arity_error x "2")
      
let (bb_typer: typer) = fun ll -> match List.flatten ll with
  | [Bool_type_eff] -> [Bool_type_eff]
  | [x1] -> (type_error [x1] "bool")
  | x    -> (arity_error x "1")

let (ii_typer: typer) = fun ll -> match List.flatten ll with
  | [Int_type_eff] -> [Int_type_eff]
  | [x1] -> (type_error [x1] "int")
  | x    -> (arity_error x "1")

let (ff_typer: typer) = fun ll -> match List.flatten ll with
  | [Real_type_eff] -> [Real_type_eff]
  | [x1] -> (type_error [x1] "real")
  | x    -> (arity_error x "1")

let (oo_typer: typer) = fun ll -> match List.flatten ll with
  | [Int_type_eff] -> [Int_type_eff]
  | [Real_type_eff] -> [Real_type_eff]
  | [x0] -> type_error [x0] "int or real"
  | x ->  arity_error x "1"

let (sf_typer: Ident.t -> typer) =
  fun id ceff_list -> 
    try let _ = float_of_string (Ident.to_string id) in
      (match ceff_list with
	 | [] -> [Real_type_eff]
	 | x -> (arity_error x "0") )
    with Failure  "float_of_string" -> 
      raise (EvalType_error(
	       Printf.sprintf "Error when conterting %s into a float"
		 (Ident.to_string id)))
	
let (si_typer: Ident.t -> typer) =
  fun id ceff_ll -> 
    try let _ = int_of_string (Ident.to_string id) in
      (match ceff_ll with
	 | [] -> [Int_type_eff]
	 | x -> (arity_error x "0"))
    with Failure  "int_of_string" -> 
      raise (EvalType_error(
	       Printf.sprintf "Error when conterting %s into an int"
		 (Ident.to_string id)))

let (sb_typer: typer) = fun ll -> match List.flatten ll with
  | [] -> [Bool_type_eff]
  | x -> (arity_error x "0")
      
let (fi_typer: typer) = fun ll -> match List.flatten ll with
  | [Real_type_eff] -> [Int_type_eff]
  | [x] -> type_error [x] "real"
  | x   -> arity_error x "1"

let (if_typer:  typer) = fun ll -> match List.flatten ll with
  | [Int_type_eff] -> [Real_type_eff]
  | [x] -> type_error [x] "int"
  | x   -> arity_error x "1"
      
let (ite_typer : typer) = function
  | [[Bool_type_eff]; t; e] -> 
      if t = e then t else 
	(type_error (List.flatten [[Bool_type_eff]; t; e]) "bool*any*any")
  | x -> (arity_error x "3")

let (boolred_typer : typer) =
  fun ceff_ll ->
    let _ = List.fold_left
      (fun acc ceff ->
	 match (ceff) with
	   | (Bool_type_eff) -> acc
	   | _ -> (type_error [ceff] "bool")
      )
      true
      (List.flatten ceff_ll)
    in
      [Bool_type_eff]
      
(* exported *)
let (type_eval: op -> typer) = function
  | TRUE_n  
  | FALSE_n -> sb_typer
  | ICONST_n id -> si_typer id
  | RCONST_n id -> sf_typer id
  | NOT_n -> bb_typer
  | REAL2INT_n -> fi_typer
  | INT2REAL_n -> if_typer
  | AND_n  
  | OR_n   
  | XOR_n  
  | IMPL_n -> bbb_typer
  | EQ_n   
  | NEQ_n  
  | LT_n   
  | LTE_n  
  | GT_n   
  | GTE_n  -> aab_typer 
  | IF_n -> ite_typer
  | UMINUS_n -> oo_typer
  | MINUS_n 
  | PLUS_n  
  | SLASH_n 
  | TIMES_n -> ooo_typer
  | IUMINUS_n  -> ii_typer
  | DIV_n  
  | MOD_n 
  | IMINUS_n 
  | IPLUS_n  
  | ISLASH_n 
  | ITIMES_n -> iii_typer 
  | RUMINUS_n -> ff_typer 
  | RMINUS_n 
  | RPLUS_n  
  | RSLASH_n 
  | RTIMES_n -> fff_typer
  | NOR_n   
  | DIESE_n -> boolred_typer

(*********************************************************************************)
   
let (bbb_evaluator:(bool -> bool -> bool) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Bool_const_eff v0; Bool_const_eff v1] -> [Bool_const_eff (op v0 v1)]
    | _ -> assert false (* should not occur because eval_type is called before *)
          
let (ooo_evaluator:(int -> int -> int) -> (float -> float -> float) -> 
      const_evaluator) =
  fun opi opr -> fun ll -> match List.flatten ll with
      | [Int_const_eff v0; Int_const_eff v1] -> [Int_const_eff (opi v0 v1)]
      | [Real_const_eff v0; Real_const_eff v1] -> [Real_const_eff (opr v0 v1)]
	  (* XXX should we evaluate reals ??? *)
      | _ -> assert false (* should not occur because eval_type is called before *)

let (iii_evaluator:(int -> int -> int) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Int_const_eff v0; Int_const_eff v1] -> [Int_const_eff (op v0 v1)]
    | _ -> assert false (* should not occur because eval_type is called before *)

let (aab_evaluator:('a -> 'a -> bool) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [v0; v1] -> [Bool_const_eff (op v0 v1)]
    | _ -> assert false (* should not occur because eval_type is called before *)

let (fff_evaluator:(float -> float -> float) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Real_const_eff v0; Real_const_eff v1] -> [Real_const_eff (op v0 v1)]
    | _ -> assert false (* should not occur because eval_type is called before *)
	
let (bb_evaluator:(bool -> bool) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Bool_const_eff v0] -> [Bool_const_eff (op v0)]
    | _ -> assert false (* should not occur because eval_type is called before *)
	
let (ii_evaluator:(int -> int) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Int_const_eff v0] -> [Int_const_eff (op v0)]
    | _ -> assert false (* should not occur because eval_type is called before *)
	
let (ff_evaluator:(float -> float) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Real_const_eff v0] -> [Real_const_eff (op v0)]
    | _ -> assert false (* should not occur because eval_type is called before *)
	
let (oo_evaluator:(int -> int) -> (float -> float) -> const_evaluator) =
  fun opi opr  -> fun ll -> match List.flatten ll with
    | [Int_const_eff v0] -> [Int_const_eff (opi v0)]
    | [Real_const_eff v0] -> [Real_const_eff (opr v0)]
	(* XXX should we evaluate reals ??? *)
    | _ -> assert false (* should not occur because eval_type is called before *)
	
let (sf_evaluator: Ident.t -> const_evaluator) =
  fun id ceff_ll -> 
    try let v = float_of_string (Ident.to_string id) in
      [Real_const_eff v]
    with Failure  "float_of_string" -> 
      raise (EvalConst_error(
	       Printf.sprintf "Error when conterting %s into a float"
		 (Ident.to_string id)))

let (si_evaluator: Ident.t -> const_evaluator) =
  fun id ceff_ll -> 
    try let v = int_of_string (Ident.to_string id) in
      [Int_const_eff v]
    with Failure  "int_of_string" -> 
      raise (EvalConst_error(
	       Printf.sprintf "Error when conterting %s into an int"
		 (Ident.to_string id)))

let (sb_evaluator: bool -> const_evaluator) =
  fun v ceff_ll -> 
      [Bool_const_eff v]

let (fi_evaluator:(float -> int) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Real_const_eff v0] -> [Int_const_eff (op v0)]
    | _ -> assert false (* should not occur because [eval_type] is called before *)

let (if_evaluator: (int -> float) -> const_evaluator) =
  fun op -> fun ll -> match List.flatten ll with
    | [Int_const_eff v0] -> [Real_const_eff (op v0)]
    | _ -> assert false (* should not occur because [eval_type] is called before *)
	
let (ite_evaluator : const_evaluator) =
  function
    | [[Bool_const_eff c]; t; e] -> if c then t else e
    | _ -> assert false (* should not occur because [eval_type] is called before *)

let (boolred_evaluator : int -> const_evaluator) =
  fun max ceff_ll ->
    let nb = List.fold_left
      (fun acc -> function
	 | (Bool_const_eff b) -> if b then acc+1 else acc | _ -> assert false)
      0
      (List.flatten ceff_ll)
    in
      [Bool_const_eff (nb <= max)]
      
	 
(* exported *)
let (const_eval: op -> const_evaluator) = 
  fun op ll -> 
    (* we first check the type so that we do not need to check it during the const
       evaluation *)
    ignore (type_eval op (List.map (List.map type_of_const_eff) ll));
    match op with
      | TRUE_n  -> sb_evaluator true ll
      | FALSE_n -> sb_evaluator false ll
      | ICONST_n id -> si_evaluator id ll
      | RCONST_n id -> sf_evaluator id ll
      | NOT_n -> bb_evaluator (not) ll
      | REAL2INT_n -> fi_evaluator int_of_float ll
      | INT2REAL_n -> if_evaluator float_of_int ll
      | AND_n  -> bbb_evaluator (&&) ll
      | OR_n   -> bbb_evaluator (||) ll
      | XOR_n  -> bbb_evaluator (<>) ll
      | IMPL_n -> bbb_evaluator (fun a b -> (not a) || b) ll
      | EQ_n   -> aab_evaluator (=) ll
      | NEQ_n  -> aab_evaluator (<>) ll 
      | LT_n   -> aab_evaluator (<) ll
      | LTE_n  -> aab_evaluator (<=) ll
      | GT_n   -> aab_evaluator (>) ll
      | GTE_n  -> aab_evaluator (>=) ll
      | DIV_n  -> iii_evaluator (/) ll
      | MOD_n  -> iii_evaluator (mod) ll
      | IF_n -> ite_evaluator ll
      | UMINUS_n -> oo_evaluator (fun x -> -x) (fun x -> -.x) ll
      | MINUS_n -> ooo_evaluator (-) (-.) ll
      | PLUS_n  -> ooo_evaluator (+) (+.) ll
      | SLASH_n -> ooo_evaluator (/) (/.) ll
      | TIMES_n -> ooo_evaluator ( * ) ( *.) ll
      | IUMINUS_n  -> ii_evaluator (fun x -> -x) ll
      | IMINUS_n -> iii_evaluator (-) ll
      | IPLUS_n  -> iii_evaluator (+) ll
      | ISLASH_n -> iii_evaluator (/) ll
      | ITIMES_n -> iii_evaluator ( * ) ll
      | RUMINUS_n -> ff_evaluator (fun x -> -.x) ll
      | RMINUS_n -> fff_evaluator (-.) ll
      | RPLUS_n  -> fff_evaluator (+.) ll
      | RSLASH_n -> fff_evaluator (/.) ll
      | RTIMES_n -> fff_evaluator ( *.) ll
      | NOR_n   -> boolred_evaluator 0 ll
      | DIESE_n -> boolred_evaluator 1 ll

(*********************************************************************************)
let finish_me msg = 
  print_string ("\n\tXXX predefSemantics.ml:"^msg^" ->  finish me!\n")

	
let (aa_clocker: clocker) =
  function
    | [clk1] -> clk1
    | _ -> finish_me "a good error msg"; assert false
	
let (aaa_clocker: clocker) =
  function
    | [clk1; clk2] -> 
	if clk1 = clk2 then clk1 else (finish_me "a good error msg"; assert false)
    | _ -> 
	finish_me "a good error msg"; assert false

   

(* This table contains the clock profile of predefined operators *)
let (clocking_tab: op -> clocker) = 
  fun op -> 
    assert false
    

(*********************************************************************************)
(*********************************************************************************)
(* 
pour evaluer l'galit, Pascal faisait comme ca (j'ai t plus (trop ?) brutal) :
      (*----------------------------
	Calcul de l'galit
	N.B. Sur les constantes abstraites
	on est trs mfiant
	N.B. Sur les types structure,
	on fait des appels rcursifs
	----------------------------*)
    let rec compute_eq 
	(args : const_eff list)
	= (
	  let rec fields_eq f0 f1 = (
	    match (f0, f1) with
	      | ([], []) -> 
		  [Bool_const_eff true]
		 
	      | ((f0,h0)::t0, (f1,h1)::t1) -> (
		  assert (f0 = f1);
		  match (compute_eq [h0;h1]) with
		      [Bool_const_eff false] -> [Bool_const_eff false]
		    | [Bool_const_eff true] -> (fields_eq t0 t1) 
		    | _ -> assert false
		)
	      | _ -> assert false
	  ) 
	  in
	    match args with
		[Bool_const_eff v0; Bool_const_eff v1] -> [Bool_const_eff (v0 = v1)]
	      |	[Int_const_eff v0; Int_const_eff v1] -> [Bool_const_eff (v0 = v1)]
	      |	[Real_const_eff v0; Real_const_eff v1] -> (
		  let res = (v0 = v1) in
		    warning src 
		      (sprintf "float in static exp: %f=%f evaluated as %b" v0 v1 res);
		    [Bool_const_eff res]
		)
		  (*
		    2007-07 obsolete

		    |  [Extern_const_eff (v0, t0); Extern_const_eff (v1, t1)] -> (
		    if (t0 <> t1) then (
		    type_error args "t*t for some type t"
		    ) else if (v0 <> v1) then (
		    uneval_error args (
		    sprintf "%s=%s (external constants)"
		    (string_of_fullid v0)
		    (string_of_fullid v1)
		    )
		    ) else (
		    [Bool_const_eff true]
		    )
		    )
		  *)
	      |	[Enum_const_eff (v0, t0); Enum_const_eff (v1, t1)] -> (
		  if (t0 = t1) then [Bool_const_eff (v0 =  v1)]
		  else type_error args "t*t for some type t"
		)
	      |	[Struct_const_eff (f0, t0); Struct_const_eff (f1, t1)] -> (
		  if (t0 = t1) then (fields_eq f0 f1)
		  else type_error args "t*t for some type t"
		)
	      |	[x;y] -> type_error args "t*t for some type t"
	      |	x -> arity_error args "2"
	) 
    in
 *)
